/**
 * This is the Message class.
 */

package salsa_lite.core.language;

import salsa_lite.core.naming.URI;
import salsa_lite.core.services.ServiceFactory;

public class Message implements java.io.Serializable {
	public URI getTargetURI() {
		if (getTarget().getUAN() != null) return getTarget().getUAN();
		else return getTarget().getUAL();
	}

	public String getTargetIdentifier() {
		return getTargetURI().toString();
	}

	/********
		*	The following satisfy the Transportable interface
	 ********/
	public int		hashCode()			{ return getTarget().hashCode(); }

	/*
	 *	messageId is a unique message generated for each
	 *	message and is used for token resolution.
	 */
	public	String		messageId;
	public	String		getMessageId()			{ return messageId; }
	public	void		setMessageId(String messageId)	{ this.messageId = messageId; }

	/**
	 *	Actor that the message is being sent from and
	 *	Actor that the message is being sent to.
	 */
	public	ActorReference	source, target;
	public	ActorReference	getSource()				{ return source; }
	public	void		setSource(ActorReference source)	{ this.source = source; }

	public	ActorReference	getTarget()				{ return target; }
	public	void		setTarget(ActorReference target)	{ this.target = target; }
	/**
	 *	Name of the method that should be invoked.
	 */
	public	String 		methodName;
	public	String 		getMethodName()				{ return methodName; }
	public	void		setMethodName(String methodName)	{ this.methodName = methodName; }

	/**
	 *	Array of values that are copied to be the arguments of the method. If one of
	 *	the arguments is the value of a yet to be evaluated token an instance of the
	 *	Token class should be subituted for it until it is evaluated.
	 */
	public	Object[] 	arguments;
	public	Object[]	getArguments()		{ return arguments; }

	/**
	 *	This is the number of tokens which must be received by the message before
	 *	it can be processed, and a vector of those tokens.
	 */
	public	int		requiredTokens = 0;
	public	int		getRequiredTokens()	{ return requiredTokens; }

	/**
	 *	After this message has been processed, it's return token is passed
	 *	to the joinDirector and the continuationToken, if they aren't null.
	 */
	public	int		joinPosition = 0;
	public	void		setJoinPosition(int joinPosition) {
		this.joinPosition = joinPosition;
	}
	public	int		getJoinPosition()	{ return joinPosition; }
	public	Token		continuationToken = null;
	public	Token		getContinuationToken()	{
		if (continuationToken != null && continuationToken.isJoinDirector()) {
			continuationToken.joinInputs = joinPosition+1;
		}
		return continuationToken;
	}

	/**
	 *	A message can be sent with a property. These methods
	 *	handle assigning of properties and their information
	 *	to a method, and are used by the compiler.  They also
	 *	provide access to this information for use by the
	 *	salsa.language package.
	 */
	public	String		propertyName = null;
	public	Object[]	propertyParameters = null;
	public	String		getProperty() { return propertyName; }
	public	Object[]	getPropertyParameters() { return propertyParameters; }

	public	void setProperty(String propertyName, Object[] propertyParameters) {
		this.propertyName = propertyName;
		this.propertyParameters = propertyParameters;

		if (propertyName.equals("waitfor")) {
			for (int i = 0; i < propertyParameters.length; i++) {
				if ( !(propertyParameters[i] instanceof Token) ) {
					System.err.println("Message Creation Error:");
					System.err.println("\tA parameter of a waitfor property was not a token.");
					System.err.println("\tOn message: " + toString());
					return;
				}
				Token currentToken = (Token)propertyParameters[i];

				if (currentToken.getValue() == null) {
					requiredTokens++;
					if (target != null) currentToken.addTarget(target, messageId, -3);
					else currentToken.addTarget(source, messageId, -3);
				}
			}
		}
	}

	/**
	 *	The process method in actor calls resolveContinuations,
	 *	with the return value for processing this message.  That
	 *	token is sent to the messages joinDirector and continuationToken's
	 *	targets, if they exist.
	 */
	public void resolveContinuations(Object value) {
		/*
		 * We need to specify where the resolve message is coming from.
		 * if this message is being resolved, that means it's at its
		 * target.
		 */
		if (continuationToken != null) {
			if (!continuationToken.isJoinDirector()) {
				continuationToken.resolve(target, value, messageId);
			} else {
				/**
				 * If this message was part of a join block, we need to also send
				 * the token to the join director, specifying where that token needs
				 * to go in the array returned by the join director.  We do this by
				 * simply sending a message to the join director.
				 */
				ActorReference joinDirectorReference = continuationToken.getJoinDirector();

				Object[] arguments = { new Integer(joinPosition), value };
				ServiceFactory.getStage().sendMessageFromLocal( new Message(target, joinDirectorReference, "resolveJoinToken", arguments, null, null, messageId + "_resolveMessage") );
			}
		}
	}

	/**
	 *	This method is called when a token has been received that this
	 *	message is waiting for.  This updates the corresponding pendingToken 
	 *	in the message.  When all pendingTokens have been resolved,
	 *	this message can be placed in it's target actors mailbox.
	 */
	public void resolveToken(Object value, int position) {
		/**
		 * Tokens are resolved according to their position,
		 * position >= 0 is that position in the array of
		 * arguments. -1 is the target actor of the message,
		 * -2 is the methodname of the message. -3 corresponds
		 * to a synchronization token, so nothing needs to be
		 * updated.
		 */

		if (position >= 0) {
			arguments[position] = value;
		} else if (position == -1) {
			if (value instanceof ActorReference) {
				target = (ActorReference)value;
			} else {
				System.err.println("Token Resolution Error:");
				System.err.println("\tA message was sent a token to resolve it's target which was");
				System.err.println("\tnot an actor.");
				System.err.println("\tOccured in message: " + toString());
			}
		} else if (position == -2) {
			if (value instanceof String) {
				methodName = (String)value;
			} else {
				System.err.println("Token Resolution Error:");
				System.err.println("\tA message was sent a token to resolve it's method name which was");
				System.err.println("\tnot a String.");
				System.err.println("\tOccured in message: " + toString());
			}
		}
		requiredTokens--;
	}

	/**
	 * These are the constructors for message.  The second is used if the message
	 * is part of a join continuation.  The constructor discovers which tokens are
	 * needed for it to be processed, and tries to resolve tokens which already have
	 * a user specified value. (ie done through code like: token x = 5;)
	 * This constructor allows for the target actor and method name to be tokens as
	 * well as the parameters.
	 */
	public Message(ActorReference source, Object target, Object methodName, Object[] arguments, Token synchronizationToken, Token continuationToken, String messageId) {
		this.messageId = messageId;
		init(source, target, methodName, arguments, synchronizationToken, continuationToken);
	}

	public Message(ActorReference source, Object target, Object methodName, Object[] arguments, Token synchronizationToken, Token continuationToken) {
		init(source, target, methodName, arguments, synchronizationToken, continuationToken);
	}

	private void init(ActorReference source, Object target, Object methodName, Object[] arguments, Token synchronizationToken, Token continuationToken) {
		if (messageId == null) {
			this.messageId = IdentifierGenerator.generateUniqueMessageId();
		}
		this.source = source;

		ActorReference targetActor;

		/**
		 *	The target actor can be either a token or an actor
		 *	reference, if it is an actor reference we do nothing
		 *	special.
		 */
		if (target instanceof ActorReference) {
			this.target = (ActorReference)target;
			targetActor = this.target;
		} else {
			Token targetToken = (Token)target;
			Object value = targetToken.getValue();
			if (value != null) this.target = (ActorReference)value;

			if (this.target == null) {
				requiredTokens++;
				/**
				 *	Here we add the source actor as the target for this token, because
				 *	when a message gets sent with an unknown target actor, it gets placed
				 *	in its source's mailbox until that actor has been resolved.
				 */
				targetActor = source;
				targetToken.addTarget( targetActor, messageId, -1 );
			} else {
				targetActor = this.target;
			}
		}

		/**
		 *	The methodName can be either a token or a string,
		 *	if it is a string we do nothing special, if it is
		 *	a token we need to wait for it to be resolved.
		 */
		if (methodName instanceof String) {
			this.methodName = (String)methodName;
		} else {
			Token methodNameToken = (Token)methodName;
			Object value = methodNameToken.getValue();
			if (value != null) this.methodName = (String)value;
			else {
				requiredTokens++;
				methodNameToken.addTarget( targetActor, messageId, -2 );
			}
		}

		/**
		 *	We must find the tokens in the arguments, and set
		 *	up some information about them, so they can be
		 *	resolved at a later time.
		 */
		for (int i = 0; i < arguments.length; i++) {
			if (arguments[i] instanceof Token) {
				Token token = (Token)arguments[i];
				arguments[i] = token.getValue();
				if (arguments[i] == null) {
					requiredTokens++;
					token.addTarget( targetActor, messageId, i );
				}
			}
		}
		this.arguments = arguments;

		/**
		 *	If there is a synchronization token, we add it to
		 *	this messages list of pending tokens as well.
		 */
		if (synchronizationToken != null && synchronizationToken.getValue() == null) {
			requiredTokens++;
			synchronizationToken.addTarget( targetActor, messageId, -3 );
		}
		/**
		 *	If the results of this message invocation need to
		 * 	be sent anywhere, the continuationToken contains
		 *	the information about this.
		 */
		this.continuationToken = continuationToken;

		if (continuationToken != null && continuationToken.isJoinDirector()) joinPosition = continuationToken.getJoinPosition();
	}

	/**
	 *	The following methods return a human readable String representation
	 *	of this message.
	 */
	public String toString() {
		String description = methodName + "(";

		for (int i = 0; i < arguments.length; i++) {
			if (arguments[i] == null) {
				description += "null";
			} else {
				description += arguments[i].getClass().getName();
			}
			if (i+1 != arguments.length) description += ", ";
		}
		description += "), target: ";

		if (target != null) description += target.toString();
		else description += "null";

		description += ", source: ";

		if (source != null) description += source.toString();
		else description += "null";

		return description;
	}
}
