package salsa_lite.core.services;

import java.lang.reflect.Constructor;

import salsa_lite.core.language.Actor;
import salsa_lite.core.language.ActorReference;
import salsa_lite.core.language.IdentifierGenerator;

import salsa_lite.core.naming.UAL;

/**
 * The ServiceFactory manufactures references to the standard location dependent
 * services used by an actor. These include the TheaterService,
 * TransportService, NamingService, and ReceptionService.
 * 
 * In the case of a standalone actor, the ServiceFactory will create a theater
 * instance if one is needed. If a theater is already running a reference to
 * will be passed to the requesting actor.
 * 
 * The ServiceFactory is implemented as a synchronized, passive, singleton.
 * 
 * The default services are:
 *	-Dtransport=wwc.messaging.WWCTransportService
 *	-Dreception=wwc.messaging.WWCReceptionService
 *	-Doutput=wwc.resources.StandardOutput
 *	-Derror=wwc.resources.StandardError
 * 
 * Classes defined in system properties must have no argument contructors and
 * should implement the their respective service interface.
 * 
 * @author stepha, deselt
 */

public class ServiceFactory {
	private static boolean local = false;
	private static int number_stages = 1;

	public static void initialize() {
		if (System.getProperty("local") != null) {
			receptionClass	= "salsa_lite.local.LocalReceptionService";
			outputClass	= "salsa_lite.local.LocalOutputServiceState";
			errorClass	= "salsa_lite.local.LocalErrorServiceState";
			stageClass	= "salsa_lite.local.LocalStageService";
			garbageClass	= "salsa_lite.local_gc.LocalGarbageService";
			transportClass	= null;
			namingClass	= null;
			getStage();
		} else if (System.getProperty("wwc") != null) {
			transportClass	= "salsa_lite.wwc.transport.WWCTransportService";
			receptionClass	= "salsa_lite.wwc.reception.WWCReceptionService";
			namingClass	= "salsa_lite.wwc.naming.WWCNamingService";
			stageClass	= "salsa_lite.wwc.stage.WWCStageService";
			outputClass	= "salsa_lite.wwc.io.WWCOutputServiceState";
			errorClass	= "salsa_lite.wwc.io.WWCErrorServiceState";
		}
	}

	// Service implementations.
	private static ReceptionService		reception	= null;
	private static TransportService		transport	= null;
	private static OutputService		output		= null;
	private static ErrorService		error		= null;
	private static NamingService		naming		= null;
	private static StageService		stage		= null;
	private static GarbageService		garbage		= null;

	//default service classes
	private static String			transportClass	= "salsa_lite.wwc.transport.WWCTransportService";
	private static String			receptionClass	= "salsa_lite.wwc.reception.WWCReceptionService";
	private static String			namingClass	= "salsa_lite.wwc.naming.WWCNamingService";
	private static String			stageClass	= "salsa_lite.wwc.stage.WWCStageService";
	private static String			outputClass	= "salsa_lite.wwc.io.WWCOutputServiceState";
	private static String			errorClass	= "salsa_lite.wwc.io.WWCErrorServiceState";
	private static String			garbageClass	= "salsa_lite.local_gc.LocalGarbageService";

	//set service classes
	public static synchronized void		setTransportClass(String className)	{ transportClass = className; }
	public static synchronized void		setReceptionClass(String className)	{ receptionClass = className; }
	public static synchronized void		setNamingClass(String className)	{ namingClass = className; }
	public static synchronized void		setStageClass(String className)		{ stageClass = className; }
	public static synchronized void		setOutputClass(String className)	{ outputClass = className; }
	public static synchronized void		setErrorClass(String className)		{ errorClass = className; }

	public static void printErrorMessage(String className, Exception e) {
		System.err.println("Service Factory error: ");
		System.err.println("\tError creating service: " + className);
		System.err.println("\tGenerated exception: " + e);
		System.err.println("\tWith message: " + e.getMessage());
		e.printStackTrace();
	}

	/**
	 * Instantiates an actor with the given className, uan and ual identifier.
	 */
	public static ActorReference createActor(String className, String identifier) {
		try {
			Class[] parTypes = { Class.forName("java.lang.String") };
			Constructor actorConstructor = Class.forName(className).getConstructor(parTypes);

			Object[] args = { identifier };
			Actor actor = (Actor)actorConstructor.newInstance( args );
			ServiceFactory.getStage().createActor(actor, getReception().getLocalHost(), getReception().getLocalPort());

			return actor.getReference();
		} catch (Exception e) {
			System.err.println("Service Factory Error:");
			System.err.println("\tCould not create system actor: " + className);
			System.err.println("\tWith identifier: " + identifier);
			System.err.println("\tThrew exception: " + e);
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Returns an implementation of the output service. If no service has been specified,
	 * it checks the system property "-Doutput=<transportservice>" and instantiates that
	 * implementation
	 * @return OutputService
	 */
	public synchronized static OutputService getOutput() {
		if (output == null) {
			// Check for the messagingService.
			String className = System.getProperty( "output" );
			if (className == null) className = outputClass;

			try {
				output = (OutputService)createActor(className, "system/StandardOutput");
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
		
		return output;
	}

	/**
	 * Returns an implementation of the error service. If no service has been specified,
	 * it checks the system property "-Derror=<errorservice>" and instantiates that
	 * implementation
	 * @return ErrorService
	 */
	public synchronized static ErrorService getError() {
		if (error == null) {
			// Check for the messagingService.
			String className = System.getProperty( "error" );
			if (className == null) className = errorClass;

			try {
				error = (ErrorService)createActor(className, "system/StandardError");
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
		
		return error;
	}

	/**
	 * Returns an implementation of the transport service. If no service has been specified,
	 * it checks the system property "-Dtransport=<transportservice>" and instantiates that
	 * implementation
	 * @return TransportService
	 */
	public synchronized static TransportService getTransport() {
		if (local) {
			return null;
		} else if (transport == null) {
			// Check for the messagingService.
			String className = System.getProperty( "transport" );
			if (className == null) className = transportClass;

			try {
				transport = (TransportService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
		
		return transport;
	}

	/**
	 * Returns the ReceptionService. If no reception service is specified,
	 * it uses the system property "-Dreception=<receptionservice>" to
	 * instantiate the given service.
	 * @return ReceptionService
	 */
	public synchronized static ReceptionService getReception() {
		if (local) {
			return null;
		} else if (reception == null) {
			// Check for the reception service.
			String className = System.getProperty( "reception" );
			if (className == null) className = receptionClass;
	
			try {
				reception = (ReceptionService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
		
		return reception;
	}

	/**
	 * Returns the namingImpl.
	 * @return NamingService
	 */
	public synchronized static NamingService getNaming() {
		if (naming == null) {
			// Check for the lookup service.
			String className = System.getProperty( "naming" );
			if (className == null) className = namingClass;

			try {
				naming = (NamingService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
			
		return naming;
	}

	/**
	 * Returns the stageImpl.
	 * @return StageService
	 */
	public synchronized static StageService getStage() {
		if (stage == null) {
			// Check for the lookup service.
			String className = System.getProperty( "stage" );
			if (className == null) className = stageClass;

			try {
				stage = (StageService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
			
		return stage;
	}

	/**
	 * Returns the stageImpl.
	 * @return StageService
	 */
	public synchronized static GarbageService getGarbage() {
		if (garbage == null) {
			// Check for the lookup service.
			String className = System.getProperty( "garbage" );
			if (className == null) className = garbageClass;

			try {
				garbage = (GarbageService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				printErrorMessage(className, e);
			}
		}
			
		return garbage;
	}
}
