package salsa_lite.wwc.naming;

import salsa_lite.core.naming.UAN;
import salsa_lite.core.naming.UAL;
import salsa_lite.core.services.NamingService;

import salsa_lite.core.naming.UANProtocol;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.Socket;
import java.util.Hashtable;

/**
 * This class implements a UANP client. After the class is constructed, several
 * threads may pass through it to conduct UANP operations through the
 * NamingService interface in salsa.language.
 * 
 * @version $Id: WWCNamingService.java,v 1.1 2007/07/31 14:07:11 deselt Exp $
 * @author Abe Stephens -- stepha@rpi.edu
 * @author Travis Desell
 * @author Carlos Varela 
 */
public class WWCNamingService implements NamingService {

	private Hashtable<String, PrintWriter> outgoingStreams;
	private Hashtable<String, BufferedReader> incomingStreams;

	public WWCNamingService() {
		outgoingStreams = new Hashtable<String, PrintWriter>();
		incomingStreams = new Hashtable<String, BufferedReader>();
	}

	public void getStreams(String host, int port) {
		try {
			Socket socket = new Socket(host, port);
			String id = host + ":" + port;
			incomingStreams.put(id, new BufferedReader(new InputStreamReader(socket.getInputStream())));
			outgoingStreams.put(id, new PrintWriter(socket.getOutputStream(), true));
		} catch (Exception e) {
			System.err.println("WWCNamingService error: getting streams to name server: " + host + ":" + port);
			System.err.println("Exception: " + e);
			e.printStackTrace();
		}
	}


	public UAL get(UAN uan) {
		String targetId = uan.getHost() + ":" + uan.getPort();

		PrintWriter out = null;
		BufferedReader in = null;
		synchronized(outgoingStreams) {
			out = outgoingStreams.get(targetId);
			if (out == null) {
				getStreams(uan.getHost(), uan.getPort());
				out = outgoingStreams.get(targetId);
			}
			in = incomingStreams.get(targetId);
		}

		if (in == null || out == null) return null;

		String request = UANProtocol.VERSION + "\n" + UANProtocol.GET_REQUEST_CODE + "\n" + uan.getIdentifier() + "\n";
		String version = null;
		String returnCode = null;
		String result = null;
		synchronized(out) {
			/********
				*	Send the request
			 ********/
			try {
				out.println(request);
				out.flush();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while sending UANP GET request: " + request);
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}

			/********
				*	Read the response
			 ********/
			try {
				version = in.readLine();
				returnCode = in.readLine();
				result = in.readLine();
				in.readLine();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while reading UANP GET response");
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}
		}

		if (!version.equals(UANProtocol.VERSION)) {
			System.err.println("WWCNamingService warning.  Different Name Server version.");
			System.err.println("Local Version: " + UANProtocol.VERSION);
			System.err.println("Remote Version: " + version);
		}
		if (!returnCode.equals(UANProtocol.FOUND_STATUS_CODE)) {
			/********
				*	UAN was not found at the name server
			 ********/
			return null;
		}

		UAL ual = null;
		try {
			ual = new UAL(result);
		} catch (Exception e) {
			System.err.println("WWCNamingService error. Name Server returned invalid UAL: " + result);
			System.err.println("Exception: " + e);
			e.printStackTrace();
			return null;
		}
		return ual;
	}

	public void put(UAN uan, UAL ual) {
		String targetId = uan.getHost() + ":" + uan.getPort();

		PrintWriter out = null;
		BufferedReader in = null;
		synchronized(outgoingStreams) {
			out = outgoingStreams.get(targetId);
			if (out == null) {
				getStreams(uan.getHost(), uan.getPort());
				out = outgoingStreams.get(targetId);
			}
			in = incomingStreams.get(targetId);
		}

		String request = UANProtocol.VERSION + "\n" + UANProtocol.PUT_REQUEST_CODE + "\n" + uan.getIdentifier() + "\n" + ual.toString();
		String version = null;
		String returnCode = null;
		String result = null;

		synchronized(out) {
			/********
				*	Send the request
			 ********/
			try {
				out.println(request);
				out.flush();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while sending UANP PUT request: " + request);
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}

			/********
				*	Read the response
			 ********/
			try {
				version = in.readLine();
				returnCode = in.readLine();
				result = in.readLine();
				in.readLine();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while reading UANP PUT response");
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}
		}

		if (!version.equals(UANProtocol.VERSION)) {
			System.err.println("WWCNamingService warning.  Different Name Server version.");
			System.err.println("Local Version: " + UANProtocol.VERSION);
			System.err.println("Remote Version: " + version);
		}
		if (!returnCode.equals(UANProtocol.MODIFY_STATUS_CODE)) {
			/********
				*	UAN was put at the name server
			 ********/
			System.err.println("WWCNamingService error. UANP PUT did not work for: " + request);
		}
	}

	public void update(UAN uan, UAL ual) {
		put(uan, ual);
	}

	public void delete(UAN uan) {
		String targetId = uan.getHost() + ":" + uan.getPort();

		PrintWriter out = null;
		BufferedReader in = null;
		synchronized(outgoingStreams) {
			out = outgoingStreams.get(targetId);
			if (out == null) {
				getStreams(uan.getHost(), uan.getPort());
				out = outgoingStreams.get(targetId);
			}
			in = incomingStreams.get(targetId);
		}

		String request = UANProtocol.VERSION + "\n" + UANProtocol.DELETE_REQUEST_CODE + "\n" + uan.getIdentifier();
		String version = null;
		String returnCode = null;
		String result = null;

		synchronized(out) {
			/********
				*	Send the request
			 ********/
			try {
				out.println(request);
				out.flush();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while sending UANP DELETE request: " + request);
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}

			/********
				*	Read the response
			 ********/
			try {
				version = in.readLine();
				returnCode = in.readLine();
				result = in.readLine();
				in.readLine();
			} catch (Exception e) {
				System.err.println("WWCNamingService error while reading UANP DELETE response");
				System.err.println("Exception: " + e);
				e.printStackTrace();
			}
		}

		if (!version.equals(UANProtocol.VERSION)) {
			System.err.println("WWCNamingService warning.  Different Name Server version.");
			System.err.println("Local Version: " + UANProtocol.VERSION);
			System.err.println("Remote Version: " + version);
		}
		if (!returnCode.equals(UANProtocol.DELETE_STATUS_CODE)) {
			/********
				*	UAN was put at the name server
			 ********/
			System.err.println("WWCNamingService error. UANP DELETE did not work for: " + request);
			System.err.println("Naming Service may not implement UANP DELETE");
		}
	}
}
